import os
from typing import Optional
from metaflow.datastore.datastore_storage import DataStoreStorage
from ..datastore.core import ObjectStorage, DatastoreInterface, STORAGE_FORMATS
from .exceptions import ModelException
from ..exceptions import KeyNotFoundError
from ..datastructures import ModelArtifact
from collections import namedtuple
from ..exceptions import KeyNotCompatibleWithObjectException
from datetime import datetime
import re

MODELS_PEFFIX = "mf.models"


ARTIFACT_STORE_NAME = "artifacts"

METADATA_STORE_NAME = "metadata"

ARTIFACT_METADATA_STORE_NAME = "artifact_metadata"


# Every model key will be prefixed with a `/models/` and will have a structure of the form:
# <root_prefix>/models/artifacts/<model-uuid>
# The root_prefix can be set at the decorator level and the /models/ is a fixed prefix that also helps distinguish between the
# different types of objects such as `checkpoints`/`models` etc.
MODELS_KEY_PATTERN = re.compile(
    r"^(?P<root_prefix>.*)/models/artifacts/(?P<model_id>.*?)$"
)

MODELS_METADATA_KEY_PATTERN = re.compile(
    r"^(?P<root_prefix>.*)/models/metadata/(?P<flow_name>.*?)/(?P<runid>.*?)/(?P<step_name>.*?)/(?P<taskid>.*?)/(?P<key_name>.*?)$"
)

ModelPathComponents = namedtuple(
    "ModelPathComponents",
    [
        "model_uuid",
        "root_prefix",
    ],
)


def __decompose_model_artifact_key(key):
    match = MODELS_KEY_PATTERN.match(key)
    if match:
        return match.groupdict()
    else:
        raise KeyNotCompatibleWithObjectException(key, "model", "Regex doesn't match")


def decompose_model_artifact_key(key):
    decomposed_key = __decompose_model_artifact_key(key)
    root_prefix, model_id = decomposed_key["root_prefix"], decomposed_key["model_id"]
    return ModelPathComponents(
        model_uuid=model_id,
        root_prefix=root_prefix,
    )


class ModelDatastore(DatastoreInterface):

    ROOT_PREFIX = MODELS_PEFFIX

    _READ_UUID = None

    MODE = None  # "READ" or "WRITE

    artifact_store: ObjectStorage = None
    metadata_store: ObjectStorage = None
    artifact_metadatastore: ObjectStorage = None  # We don't need this store for now

    def save(
        self,
        artifact: ModelArtifact,
        file_path,
        storage_format=STORAGE_FORMATS.TAR,
    ):
        save_func = None
        if storage_format == STORAGE_FORMATS.TAR:
            save_func = self.artifact_store._save_tarball
        elif storage_format == STORAGE_FORMATS.FILES:
            save_func = self.artifact_store._save_objects
        else:
            raise ModelException(
                "Incompatible storage format %s for `current.model.save`"
                % (storage_format)
            )

        full_object_url, key_path, file_size = save_func(artifact.uuid, file_path)
        _metadata = artifact.to_dict()
        _metadata.update(
            {
                "size": file_size,
                "url": full_object_url,
                "key": key_path,
                "created_on": datetime.now().isoformat(),
            }
        )
        self.save_metadata(
            attempt=artifact.attempt,
            model_id=artifact.uuid,
            metadata=_metadata,
        )
        return ModelArtifact.from_dict(_metadata)

    def load_metadata(
        self,
        model_id,
    ):
        return self.artifact_metadatastore._load_metadata(
            self.artifact_metadatastore.create_key_name(model_id, "metadata")
        )

    def save_metadata(
        self,
        attempt,
        model_id,
        metadata,
    ):
        self.metadata_store._save_metadata(
            self.metadata_store.create_key_name(attempt, model_id, "metadata"), metadata
        )
        self.artifact_metadatastore._save_metadata(
            self.artifact_metadatastore.create_key_name(model_id, "metadata"),
            metadata,
        )

    def load(
        self,
        model_id,
        path,
    ):
        try:
            metadata = self.artifact_metadatastore._load_metadata(
                self.artifact_metadatastore.create_key_name(model_id, "metadata")
            )
            storage_format = metadata.get("storage_format")
            if storage_format is None:
                raise ModelException(
                    "Malformed metadata for model with id %s." % model_id
                )
        except KeyNotFoundError:
            raise ModelException(
                "Metadata related to model with id %s not found." % model_id
            )
        load_func = None
        if storage_format == STORAGE_FORMATS.TAR:
            load_func = self.artifact_store._load_tarball
        elif storage_format == STORAGE_FORMATS.FILES:
            load_func = self.artifact_store._load_objects
        else:
            raise ModelException(
                "Unsupported storage format for model with id %s." % model_id
            )
        load_func(
            model_id,
            path,
        )

    def list(self, *args, **kwargs):
        # It will list only based on task since all models
        # are stored under one prefix.
        raise NotImplementedError

    @classmethod
    def init_read_store(
        cls,
        storage_backend: DataStoreStorage,
        pathspec: Optional[str] = None,
        attempt: Optional[str] = None,
        model_key=None,
        *args,
        **kwargs,
    ):
        datastore = cls()
        if pathspec is not None:
            if not all([pathspec, attempt]):
                raise ValueError("pathspec and attempt are required")

            datastore.metadata_store = ObjectStorage(
                storage_backend,
                root_prefix=cls.ROOT_PREFIX,
                path_components=["models", METADATA_STORE_NAME] + pathspec.split("/"),
            )
            datastore.artifact_store = ObjectStorage(
                storage_backend,
                root_prefix=cls.ROOT_PREFIX,
                path_components=["models", ARTIFACT_STORE_NAME],
            )
            datastore.artifact_metadatastore = ObjectStorage(
                storage_backend,
                root_prefix=cls.ROOT_PREFIX,
                path_components=["models", ARTIFACT_METADATA_STORE_NAME],
            )

        elif model_key is not None:
            model_path_decomp = decompose_model_artifact_key(model_key)
            datastore.artifact_store = ObjectStorage(
                storage_backend,
                root_prefix=model_path_decomp.root_prefix,
                path_components=["models", ARTIFACT_STORE_NAME],
            )
            datastore.artifact_metadatastore = ObjectStorage(
                storage_backend,
                root_prefix=model_path_decomp.root_prefix,
                path_components=["models", ARTIFACT_METADATA_STORE_NAME],
            )
            datastore._READ_UUID = model_path_decomp.model_uuid
        datastore.MODE = "READ"
        return datastore

    @classmethod
    def decompose_key(cls, key):
        return decompose_model_artifact_key(key)

    @classmethod
    def init_write_store(
        cls, storage_backend: DataStoreStorage, pathspec: str, attempt, *args, **kwargs
    ):
        datastore = cls()
        if any([pathspec is None, attempt is None]):
            raise ValueError("pathspec and attempt are required")

        datastore.metadata_store = ObjectStorage(
            storage_backend,
            root_prefix=cls.ROOT_PREFIX,
            path_components=["models", METADATA_STORE_NAME] + pathspec.split("/"),
        )
        datastore.artifact_store = ObjectStorage(
            storage_backend,
            root_prefix=cls.ROOT_PREFIX,
            path_components=["models", ARTIFACT_STORE_NAME],
        )
        datastore.artifact_metadatastore = ObjectStorage(
            storage_backend,
            root_prefix=cls.ROOT_PREFIX,
            path_components=["models", ARTIFACT_METADATA_STORE_NAME],
        )
        datastore.MODE = "WRITE"
        return datastore



from collections import namedtuple
from hashlib import sha256

import os


from metaflow import current
from metaflow.exception import MetaflowException
from typing import List, Union, Optional, TYPE_CHECKING
import os
from .exceptions import CheckpointNotAvailableException, CheckpointException
from ..utils import flowspec_utils
from .checkpoint_storage import CheckpointDatastore
from ..datastructures import CheckpointArtifact
from ..datastore.task_utils import (
    init_datastorage_object,
    resolve_storage_backend as resolve_task_storage_backend,
)
from .constants import (
    CHECKPOINT_TAG_PREFIX,
    MAX_HASH_LEN,
    DEFAULT_NAME,
    CHECKPOINT_UID_ENV_VAR_NAME,
    DEFAULT_STORAGE_FORMAT,
)

if TYPE_CHECKING:
    import metaflow

MAX_HASH_LEN = 12


def _coalesce(*args):
    return next((x for x in args if x is not None), None)


def _coalesce_lambdas(*args):
    for x in args:
        _val = x()
        if _val is not None:
            return _val
    return None


class Checkpointer:
    # TODO : this abstraction is not as well designed as it should be
    # TODO : Figure better abstraction interplay.

    _checkpoint_uid: str

    _checkpoint_datastore: CheckpointDatastore

    @property
    def current_version(self):
        return self._current_version

    def _set_current_version(self, version_id):
        self._current_version = version_id

    def set_root_prefix(self, root_prefix):
        self._checkpoint_datastore.set_root_prefix(root_prefix)

    def __init__(
        self,
        datastore: CheckpointDatastore,
        attempt: int = 0,
        # TODO: [POST RELEASE]: See if attempt can be moved elsewhere
    ) -> None:
        self._checkpoint_datastore = datastore
        self._attempt = attempt
        self._current_version = 0

    def _get_latest_checkpoint(self, safe=True):
        return self._checkpoint_datastore.latest(safe=safe)

    def _update_version(self):
        self._current_version += 1

    def load_metadata(self, version_id: int = None, name=DEFAULT_NAME) -> dict:
        return self._checkpoint_datastore.load_metadata(
            attempt=self._attempt,
            version_id=version_id,
            name=name,
        )

    def save(
        self,
        path: str,
        metadata={},
        latest=True,
        name=DEFAULT_NAME,
        storage_format=DEFAULT_STORAGE_FORMAT,
    ) -> CheckpointArtifact:
        _art = self._checkpoint_datastore.save(
            path,
            attempt=self._attempt,
            version_id=self._current_version,
            name=name,
            metadata=metadata,
            set_latest=latest,
            storage_format=storage_format,
        )
        self._update_version()
        return _art

    @classmethod
    def _from_checkpoint_and_storage_backend(
        cls, checkpoint: CheckpointArtifact, storage_backend
    ):
        """
        Used by: `Factory` to load the artifact from native MetaflowArtifactReference objects.
        """
        _checkpoint_datastore = CheckpointDatastore.init_read_store(
            storage_backend, checkpoint_key=checkpoint.key
        )
        obj = cls(
            datastore=_checkpoint_datastore,
            attempt=checkpoint.attempt,
        )
        return obj

    @classmethod
    def _from_key(cls, key: str):
        key_decomp = CheckpointDatastore.decompose_key(key)
        datastore = ReadResolver.from_key(key)
        obj = cls(
            datastore=datastore,
            attempt=key_decomp.attempt,
        )
        obj._set_current_version(key_decomp.version_id)
        return obj

    @classmethod
    def _from_checkpoint(cls, checkpoint: Union[CheckpointArtifact, dict]):
        key, pathspec = None, None
        _chckpt: CheckpointArtifact = CheckpointArtifact.hydrate(checkpoint)
        key, pathspec = _chckpt.key, _chckpt.pathspec
        # TODO [POST-RELEASE]: Suport out of task checkpoints
        datastore = ReadResolver.from_key_and_pathspec(pathspec, key)
        obj = cls(
            datastore=datastore,
            attempt=_chckpt.attempt,
        )
        obj._set_current_version(_chckpt.version_id)
        return obj

    def _load_checkpoint(
        self,
        local_path: str,
        version_id: int = None,
        name=DEFAULT_NAME,
        storage_format=None,  # Loading should require an explicit format
    ):
        self._checkpoint_datastore.load(
            local_path,
            version_id=version_id,
            attempt=self._attempt,
            name=name,
            storage_format=storage_format,
        )

    def _list_checkpoints(
        self,
        name=DEFAULT_NAME,
        attempt=None,
        within_task=True,
    ):
        return self._checkpoint_datastore.list(
            name=name, attempt=attempt, within_task=within_task
        )


class CheckpointLoadPolicy:
    @classmethod
    def fresh(
        cls,
        datastore: CheckpointDatastore,
        flow: "metaflow.FlowSpec",
    ) -> Union[CheckpointArtifact, None]:
        """
        ```python
        @checkpoint(load_policy="fresh")
        ```
        While in `fresh` mode, we want to load the "latest" checkpoint from
        what ever task is executing at the current memoment.

        The behavior is will be such that 1st attempt of any task will not load
        any checkpoint and there after it will load the checkpoint from the previous
        attempt of the task (ala the lastest checkpoint within the task).


        """
        latest_task_checkpoint_lambda = lambda: datastore.latest(
            current_task=True, safe=True
        )
        return latest_task_checkpoint_lambda()

    @classmethod
    def eager(
        cls,
        datastore: CheckpointDatastore,
        flow: "metaflow.FlowSpec",
    ) -> Union[CheckpointArtifact, None]:
        """
        ```python
        @checkpoint(load_policy="eager")
        ```
        While in `eager` mode, we want to load the "latest" checkpoint ever
        written at the "scope" level based on the kind of task that is executing.

        Setting this mode helps "checkpoints leak across executions" for the same task
        there by allowing a way to reboot the state when new executions start.
        """
        latest_task_checkpoint_lambda = lambda: datastore.latest(
            current_task=True, safe=True
        )
        latest_scope_checkpoint_lambda = lambda: datastore.latest(
            current_task=False, safe=True
        )
        return _coalesce_lambdas(
            latest_task_checkpoint_lambda,
            latest_scope_checkpoint_lambda,
        )


class ScopeResolver:
    @classmethod
    def from_namespace(cls):
        from metaflow import get_namespace

        ns = get_namespace()
        if ns is None:
            raise MetaflowException("Cannot resolve checkpoint path without Namespace.")
        return sha256(ns.encode()).hexdigest()[:MAX_HASH_LEN]

    @classmethod
    def from_tags(cls, tags):
        from metaflow import Run

        return sha256(tags[0].split(":")[1].encode()).hexdigest()[:MAX_HASH_LEN]


class ReadResolver:
    """
    Responsible for instantiating the `CheckpointDatastore` during read operations
    based on different context's.
    """

    @classmethod
    def from_key(cls, checkpoint_key):
        """"""
        storage_backend = init_datastorage_object()
        _checkpoint_datastore = CheckpointDatastore.init_read_store(
            storage_backend, checkpoint_key=checkpoint_key
        )
        return _checkpoint_datastore

    @classmethod
    def from_key_and_pathspec(cls, pathspec, checkpoint_key):
        """ """
        storage_backend = resolve_task_storage_backend(pathspec=pathspec)
        _checkpoint_datastore = CheckpointDatastore.init_read_store(
            storage_backend, checkpoint_key=checkpoint_key
        )
        return _checkpoint_datastore

    @classmethod
    def from_key_and_run(cls, run, checkpoint_key) -> CheckpointDatastore:
        """ """
        storage_backend = flowspec_utils.resolve_storage_backend(
            run=run,
        )
        _checkpoint_datastore = CheckpointDatastore.init_read_store(
            storage_backend, checkpoint_key=checkpoint_key
        )
        return _checkpoint_datastore


class WriteResolver:
    """
    Responsible for instantiating the `CheckpointDatastore` and the subsequent
    `_checkpointer_uid` which can instantiate the `Checkpointer` object outside
    of the metaflow context.
    """

    ResolverInfo = namedtuple(
        "ResolverInfo", ["flow", "run", "step", "taskid", "taskidf", "scope", "attempt"]
    )

    @classmethod
    def can_resolve_from_envionment(cls):
        if CHECKPOINT_UID_ENV_VAR_NAME in os.environ:
            return True
        return False

    @classmethod
    def decompose_checkpoint_id(cls, checkpoint_id):
        if len(checkpoint_id.split("/")) != 7:
            raise CheckpointException(
                "`%s` environment variable is not in correct format."
                % CHECKPOINT_UID_ENV_VAR_NAME
            )

        flow, run, step, taskid, taskidf, scope, attempt = checkpoint_id.split("/")
        return cls.resolver_info(flow, run, step, taskid, taskidf, scope, attempt)

    @classmethod
    def resolver_info(cls, flow, run, step, taskid, taskidf, scope, attempt):
        return cls.ResolverInfo(flow, run, step, taskid, taskidf, scope, attempt)

    @classmethod
    def construct_checkpoint_id(cls, resolver_info: ResolverInfo):
        return "/".join([str(x) for x in resolver_info])

    @classmethod
    def from_environment(
        cls,
    ):
        # Resolve the full checkpoint datastore from the environment.
        chckpt_uid = os.environ[CHECKPOINT_UID_ENV_VAR_NAME]
        _resolver_info = cls.decompose_checkpoint_id(chckpt_uid)
        pathspec = "%s/%s/%s/%s" % (
            _resolver_info.flow,
            _resolver_info.run,
            _resolver_info.step,
            _resolver_info.taskid,
        )
        storage_backend = resolve_task_storage_backend(pathspec=pathspec)
        _checkpoint_datastore = CheckpointDatastore.init_write_store(
            storage_backend,
            pathspec=pathspec,
            scope=_resolver_info.scope,
            task_identifier=_resolver_info.taskidf,
        )
        return _checkpoint_datastore, _resolver_info

    @classmethod
    def from_run(
        cls,
        run: "metaflow.FlowSpec",
        scope: str,
        task_identifier: Optional[str] = None,
        gang_scheduled_task=False,
    ):
        """
        The task-identifier gets computed in the Metaflow main process with the
        i.e. in the decorator. The pathspec we choose to write the metadata store
        depends on if the task is being gang scheduled or not.
        """

        storage_backend = flowspec_utils.resolve_storage_backend(
            run=run,
        )
        identifier = task_identifier
        resolved_pathspec_info = flowspec_utils.resolve_pathspec_for_flowspec(
            run,
        )
        # For gang scheduled tasks with @parallel decorator, all worker
        # tasks will be writing to the pathspec as the control task for
        # the metadata store.
        # The `task_identifier` for all workers tasks in @parallel
        # will be the same as the control task identifier.
        pathspec = resolved_pathspec_info.pathspec
        if gang_scheduled_task:
            pathspec = resolved_pathspec_info.control_task_pathspec

        _checkpoint_datastore = CheckpointDatastore.init_write_store(
            storage_backend,
            pathspec=pathspec,
            scope=scope,
            task_identifier=identifier,
        )

        flow_name, run_id, step_name, task_id = pathspec.split("/")
        return _checkpoint_datastore, cls.resolver_info(
            flow_name,
            run_id,
            step_name,
            task_id,
            identifier,
            scope,
            resolved_pathspec_info.current_attempt,
        )


class CheckpointReferenceResolver:
    """
    Resolve the Metaflow checkpoint object based on the flow artifact reference
    or key; Used for lineage derivation.
    """

    @classmethod
    def from_key(cls, flow, checkpoint_key):
        """
        Used by lineage derivation
        """
        checkpoint_datastore = ReadResolver.from_key_and_run(
            run=flow, checkpoint_key=checkpoint_key
        )
        key_comps = checkpoint_datastore.decompose_key(checkpoint_key)
        _checkpoint = checkpoint_datastore.load_metadata(
            attempt=key_comps.attempt,
            version_id=key_comps.version_id,
            name=key_comps.name,
        )
        if _checkpoint is None:
            raise CheckpointNotAvailableException(
                "Checkpoint with key `%s` not found." % checkpoint_key
            )
        return CheckpointArtifact.hydrate(_checkpoint)



from metaflow.exception import MetaflowException


class LoadingException(MetaflowException):
    headline = "Exception Loading Models/Checkpoints"

    def __init__(self, message):
        super(LoadingException, self).__init__(message)


class ModelException(MetaflowException):
    headline = "@model Exception"

    def __init__(self, message):
        super(ModelException, self).__init__(message)



from datetime import datetime
from typing import List, TYPE_CHECKING
from metaflow.cards import Table, Markdown, MetaflowCardComponent, Artifact
from metaflow.plugins.cards.card_modules.basic import DagComponent, SectionComponent

if TYPE_CHECKING:
    from ...datastructures import CheckpointArtifact as Checkpoint


# Helper function to format datetime
def format_datetime(iso_str):
    # Parse the ISO 8601 string to a datetime object
    dt = datetime.fromisoformat(iso_str)
    # Format the datetime object to a more readable string
    return dt.strftime("%B %d, %Y, %H:%M:%S")


def _current_task_info():
    from metaflow import current

    name_descrip = "%s [attempt: %s]" % (current.pathspec, current.retry_count)
    task_info = Markdown("Current Task : **%s**" % name_descrip)
    return task_info


def null_card(load_policy):

    return [
        Markdown("## Loaded Checkpoint Info [No checkpoint was loaded]"),
        Markdown(f"**Load Policy**: {load_policy}"),
    ]


# Function to create a lineage table
def construct_lineage_table(lineage):
    headers = ["Index", "Pathspec", "Created On", "Name", "metadata"]
    rows = []
    for idx, checkpoint in enumerate(lineage):
        created_on = (
            format_datetime(checkpoint.created_on) if checkpoint.created_on else "N/A"
        )
        rows.append(
            [
                str(k)
                for k in [
                    idx,
                    checkpoint.pathspec,
                    created_on,
                    checkpoint.name,
                ]
            ]
            + [Artifact(checkpoint.metadata)]
        )
    return Table(headers=headers, data=rows)


def create_checkpoint_card(
    loaded_checkpoint: "Checkpoint",
    checkpoint_lineage: List["Checkpoint"],
    load_policy: str,
) -> List[MetaflowCardComponent]:

    # Function to create a human-readable table from a checkpoint
    def checkpoint_info_table(checkpoint):
        info_keys = ["name", "url", "key", "pathspec", "attempt", "created_on"]
        rows = []
        for key in info_keys:
            value = getattr(checkpoint, key)
            if key == "created_on" and value:
                value = format_datetime(value)
            rows.append([Markdown("**%s**" % key.capitalize()), Markdown(str(value))])
        return Table(data=rows, headers=["Property", "Value"])

    # Function to create a table from metadata dictionary
    def construct_metadata_table(metadata):
        md_table = []
        for k, v in metadata.items():
            md_table.append([Markdown(f"**{k}**"), Artifact(v)])
        return Table(headers=["", ""], data=md_table)

    # Create sections of the card
    loaded_info_md = Markdown("## Loaded Checkpoint Info")
    loaded_info_table = checkpoint_info_table(
        loaded_checkpoint,
    )
    card_components = [
        loaded_info_md,
        Markdown(f"**Load Policy**: {load_policy}"),
        loaded_info_table,
    ]
    if loaded_checkpoint.metadata and len(loaded_checkpoint.metadata) > 0:
        card_components.extend(
            [
                Markdown("### Metadata"),
                construct_metadata_table(loaded_checkpoint.metadata),
            ]
        )

    # Construct and return the Metaflow Card
    return card_components


# Adjust the datetime parsing and formatting as necessary based on the actual



from ..datastore.core import STORAGE_FORMATS

DEFAULT_NAME = "mfchckpt"
MAX_HASH_LEN = 12
CHECKPOINTS_STORAGE_PREFIX = "mf.checkpoints"
CHECKPOINT_TASK_IDENTIFIER_ENV_VAR_NAME = "METAFLOW_CHECKPOINT_TASK_IDENTIFIER"
CHECKPOINT_UID_ENV_VAR_NAME = "METAFLOW_CHECKPOINT_UID"

TASK_CHECKPOINTS_ARTIFACT_NAME = "_task_checkpoints"
TASK_LATEST_CHECKPOINT_ARTIFACT_NAME = "_task_latest_checkpoint"

CHECKPOINT_TAG_PREFIX = "checkpoint:"

DEFAULT_STORAGE_FORMAT = STORAGE_FORMATS.FILES



import hashlib
import time
import shutil
import json
from ..checkpoints.decorator import (
    CheckpointDecorator,
    CurrentCheckpointer,
    warning_message,
)
import sys
import os
import tempfile
from metaflow.metadata_provider import MetaDatum

HUGGINGFACE_HUB_ROOT_PREFIX = "mf.huggingface_hub"


def get_tqdm_class():
    from tqdm.std import tqdm as std_tqdm

    class TqdmExt(std_tqdm):
        def __init__(self, *args, **kwargs):
            kwargs["file"] = sys.stdout
            kwargs["desc"] = (
                "[@huggingface_hub][HF-Download]"
                if not kwargs.get("desc", None)
                else "[@huggingface_hub][HF-Download] " + kwargs["desc"]
            )
            kwargs["leave"] = False
            super().__init__(*args, **kwargs)

    return TqdmExt


def show_progress():
    import logging
    from tqdm.contrib.logging import tqdm_logging_redirect

    tqdm_logger = logging.getLogger("tqdm")
    tqdm_logger.setLevel(logging.INFO)
    return tqdm_logging_redirect()


def download_model_from_huggingface(**kwargs):
    import huggingface_hub
    from glob import glob
    import os

    try:
        kwargs.pop("tqdm_class", None)
        with show_progress():
            huggingface_hub.snapshot_download(**kwargs, tqdm_class=get_tqdm_class())
    except Exception as e:
        raise e


class HuggingfaceRegistry:
    """
    This object provides syntactic sugar
    over [huggingface_hub](https://github.com/huggingface/huggingface_hub)'s
    [snapshot_download](https://huggingface.co/docs/huggingface_hub/main/en/package_reference/file_download#huggingface_hub.snapshot_download) function.
    The `current.huggingface_hub.snapshot_download` function downloads objects from huggingface hub and saves them to the Metaflow's datastore under the
    `<repo_type>/<repo_id>` name. The `repo_type` is by default `model` and can be overriden by passing the `repo_type` parameter to the `snapshot_download` function.
    """

    _checkpointer: CurrentCheckpointer = None
    _loaded_models = None

    def __init__(self, logger) -> None:
        self._logger = logger

    def _set_checkpointer(self, checkpointer: CurrentCheckpointer, temp_dir_root=None):
        self._checkpointer = checkpointer
        self._checkpointer._default_checkpointer._checkpointer.set_root_prefix(
            HUGGINGFACE_HUB_ROOT_PREFIX
        )
        self._loaded_models = HuggingfaceLoadedModels(
            checkpointer=self, logger=self._logger, temp_dir_root=temp_dir_root
        )

    @property
    def loaded(self):
        return self._loaded_models

    def _cache_name(self, name):
        return hashlib.md5(name.encode()).hexdigest()[:10]

    def _warn(self, message):
        warning_message(
            message,
            self._logger,
            prefix="[@huggingface_hub]",
        )

    def _load_or_cache_model(self, **kwargs) -> dict:
        from metaflow import current

        repo_name = kwargs["repo_id"]
        repo_type = kwargs.get("repo_type", "model")
        force_download = kwargs.get("force_download", False)
        chckpt_name = self._cache_name("{}/{}".format(repo_type, repo_name))
        chckpts = list(self._checkpointer.list(name=chckpt_name, within_task=False))
        if len(chckpts) > 0 and not force_download:
            return chckpts[0]

        _kwargs = kwargs.copy()
        _kwargs["local_dir"] = self._checkpointer.directory
        _kwargs["local_dir_use_symlinks"] = False
        start_time = time.time()
        self._warn(
            "Downloading %s from huggingface to path %s"
            % (repo_name, _kwargs["local_dir"]),
        )
        download_model_from_huggingface(**_kwargs)
        download_completion_time = time.time()
        download_time = str(round(download_completion_time - start_time, 2))
        self._warn(
            "Downloaded %s from huggingface in %s seconds. Saving checkpoint to datastore."
            % (repo_name, download_time),
        )
        chckpt_ref = self._checkpointer.save(
            name=chckpt_name,
            metadata={
                "repo_id": repo_name,
                "registry": "huggingface",
                "repo_type": repo_type,
            },
            # We set this statically to files here because
            # it will be a lot more performant than tar mode.
            storage_format="files",
        )
        _save_time = str(round(time.time() - download_completion_time, 2))
        self._warn(
            "huggingface checkpoint for %s saved to datastore in %s seconds"
            % (repo_name, _save_time),
        )
        # wipe the directory so that it's unpolluted for another function call.
        shutil.rmtree(self._checkpointer.directory)
        return chckpt_ref

    def snapshot_download(self, **kwargs) -> dict:
        """
        Downloads a model from huggingface hub and cache's it to the Metaflow's datastore.
        It passes down all the parameters to the `huggingface_hub.snapshot_download` function.

        Returns
        -------
        dict
            A reference to the artifact that was saved/retrieved from the Metaflow's datastore.
        """
        if "repo_id" not in kwargs:
            raise ValueError("repo_id is required for snapshot_download")
        return self._load_or_cache_model(**kwargs)


class HuggingfaceLoadedModels:
    """Manages loaded Hugging Face models and provides access to their local paths.

    This class is accessed through `current.huggingface_hub.loaded` and provides a dictionary-like
    interface to access the local paths of the huggingface repos specified in the `load` argument of the `@huggingface_hub` decorator.

    Usage:
    ------
    ```python
    # Basic loading and access
    @huggingface_hub(load=["mistralai/Mistral-7B-Instruct-v0.1"])
    @step
    def my_step(self):
        # Access the local path of a loaded model
        model_path = current.huggingface_hub.loaded["mistralai/Mistral-7B-Instruct-v0.1"]

        # Check if a model is loaded
        if "mistralai/Mistral-7B-Instruct-v0.1" in current.huggingface_hub.loaded:
            print("Model is loaded!")

    # Custom path and advanced loading
    @huggingface_hub(load=[
        ("mistralai/Mistral-7B-Instruct-v0.1", "/custom/path"),  # Specify custom path
        {
            "repo_id": "org/model-name",
            "force_download": True,  # Force fresh download
            "repo_type": "dataset"   # Load dataset instead of model
        }
    ])
    @step
    def another_step(self):
        # Models are available at specified paths
        pass
    ```
    """

    def __init__(
        self, checkpointer: "HuggingfaceRegistry", logger, temp_dir_root=None
    ) -> None:
        from metaflow import current

        self._namespace = current.namespace
        self._checkpointer = checkpointer
        self._logger = logger
        self._loaded_models = {}
        self._loaded_model_info = {}
        self._temp_directories = {}
        self._temp_dir_root = temp_dir_root

    def _warn(self, message):
        warning_message(
            message,
            self._logger,
            prefix="[@huggingface_hub]",
        )

    def _get_or_create_model_path(self, repo_id, chckpt_name, path=None):
        """Handle model path creation/retrieval logic"""
        if path is not None:
            if not os.path.exists(path):
                os.makedirs(path, exist_ok=True)
            return path

        # Create temp directory for model
        if self._temp_dir_root is not None:
            if not os.path.exists(self._temp_dir_root):
                os.makedirs(self._temp_dir_root, exist_ok=True)

        self._temp_directories[repo_id] = tempfile.TemporaryDirectory(
            dir=self._temp_dir_root, prefix=f"metaflow_hf_{chckpt_name}_"
        )
        return self._temp_directories[repo_id].name

    def _download_and_cache_model(self, repo_id, repo_type, path, **kwargs):
        """Download model from HF Hub and cache in datastore"""
        chckpt_name = self._checkpointer._cache_name("{}/{}".format(repo_type, repo_id))
        self._warn(
            "Downloading %s from huggingface to path %s" % (repo_id, path),
        )

        # Download directly to specified path
        kwargs["local_dir"] = path
        kwargs["local_dir_use_symlinks"] = False
        download_model_from_huggingface(repo_id=repo_id, repo_type=repo_type, **kwargs)

        self._warn(
            "Cached %s in datastore for namespace %s" % (repo_id, self._namespace),
        )
        # Save/cache in datastore
        chckpt_ref = self._checkpointer._checkpointer.save(
            name=chckpt_name,
            metadata={
                "repo_id": repo_id,
                "registry": "huggingface",
                "repo_type": repo_type,
                "kwargs": kwargs,
            },
            path=path,
            storage_format="files",
        )
        return chckpt_ref

    def _load_from_datastore(self, chckpt_ref, path):
        """Load model from datastore to specified path"""
        self._warn(
            "Loading model from datastore to %s. Model being loaded: %s"
            % (path, chckpt_ref["key"])
        )
        self._checkpointer._checkpointer.load(chckpt_ref, path=path)

    def _load_model(self, repo_id, path=None, repo_type="model", **kwargs):
        """
        Load a model from either the datastore or Hugging Face Hub.

        Args:
            repo_id (str): The Hugging Face model repo ID
            path (str, optional): Specific path to load the model into
            repo_type (str, optional): Type of repo (model/dataset)
            **kwargs: Additional arguments passed to snapshot_download
        """
        chckpt_name = self._checkpointer._cache_name("{}/{}".format(repo_type, repo_id))
        chckpts = list(
            self._checkpointer._checkpointer.list(name=chckpt_name, within_task=False)
        )

        # Setup model path and load if needed
        model_path = self._get_or_create_model_path(repo_id, chckpt_name, path=path)
        # Get or download model reference
        if len(chckpts) == 0 or kwargs.get("force_download", False):
            self._warn(
                f"Model {repo_id} not found in datastore, downloading from HuggingFace Hub"
            )
            chckpt_ref = self._download_and_cache_model(
                repo_id, repo_type, model_path, **kwargs
            )
        else:  # This means that more than 1 checkpoint exists
            chckpt_ref = chckpts[0]
            self._load_from_datastore(chckpt_ref, model_path)

        # Update tracking
        self._loaded_models[repo_id] = model_path
        self._loaded_model_info[repo_id] = chckpt_ref
        return model_path

    def __getitem__(self, key):
        if key not in self._loaded_models:
            raise KeyError(f"Model {key} not found in loaded models")
        return self._loaded_models[key]

    def __contains__(self, key):
        return key in self._loaded_models

    @property
    def info(self):
        return self._loaded_model_info

    def cleanup(self):
        for tempdir in self._temp_directories.values():
            tempdir.cleanup()


class HuggingfaceHubDecorator(CheckpointDecorator):
    """
    Decorator that helps cache, version and store models/datasets from huggingface hub.

    Parameters
    ----------
    temp_dir_root : str, optional
        The root directory that will hold the temporary directory where objects will be downloaded.

    load: Union[List[str], List[Tuple[Dict, str]], List[Tuple[str, str]], List[Dict], None]
        The list of repos (models/datasets) to load.

        Loaded repos can be accessed via `current.huggingface_hub.loaded`. If load is set, then the following happens:

        - If repo (model/dataset) is not found in the datastore:
            - Downloads the repo from Hugging Face Hub to a temporary directory (or uses specified path) for local access
            - Stores it in Metaflow's datastore (s3/gcs/azure etc.) with a unique name based on repo_type/repo_id
                - All HF models loaded for a `@step` will be cached separately under flow/step/namespace.

        - If repo is found in the datastore:
            - Loads it directly from datastore to local path (can be temporary directory or specified path)


    MF Add To Current
    -----------------
    huggingface_hub -> metaflow_extensions.obcheckpoint.plugins.machine_learning_utilities.hf_hub.decorator.HuggingfaceRegistry
        The `@huggingface_hub` injects a `huggingface_hub` object into the `current` object. This object provides syntactic sugar
        over [huggingface_hub](https://github.com/huggingface/huggingface_hub)'s
        [snapshot_download](https://huggingface.co/docs/huggingface_hub/main/en/package_reference/file_download#huggingface_hub.snapshot_download) function.
        The `current.huggingface_hub.snapshot_download` function downloads objects from huggingface hub and saves them to the Metaflow's datastore under the
        `<repo_type>/<repo_id>` name. The `repo_type` is by default `model` and can be overriden by passing the `repo_type` parameter to the `snapshot_download` function.


        Usage:
        ------

        **Usage: creating references of models from huggingface that may be loaded in downstream steps**
        ```python
            @huggingface_hub
            @step
            def pull_model_from_huggingface(self):
                # `current.huggingface_hub.snapshot_download` downloads the model from the Hugging Face Hub
                # and saves it in the backend storage based on the model's `repo_id`. If there exists a model
                # with the same `repo_id` in the backend storage, it will not download the model again. The return
                # value of the function is a reference to the model in the backend storage.
                # This reference can be used to load the model in the subsequent steps via `@model(load=["llama_model"])`

                self.model_id = "mistralai/Mistral-7B-Instruct-v0.1"
                self.llama_model = current.huggingface_hub.snapshot_download(
                    repo_id=self.model_id,
                    allow_patterns=["*.safetensors", "*.json", "tokenizer.*"],
                )
                self.next(self.train)
        ```

        **Usage: loading models directly from huggingface hub or from cache (from metaflow's datastore)**
        ```python
            @huggingface_hub(load=["mistralai/Mistral-7B-Instruct-v0.1"])
            @step
            def pull_model_from_huggingface(self):
                path_to_model = current.huggingface_hub.loaded["mistralai/Mistral-7B-Instruct-v0.1"]
        ```

        ```python
            @huggingface_hub(load=[("mistralai/Mistral-7B-Instruct-v0.1", "/my-directory"), ("myorg/mistral-lora, "/my-lora-directory")])
            @step
            def finetune_model(self):
                path_to_model = current.huggingface_hub.loaded["mistralai/Mistral-7B-Instruct-v0.1"]
                # path_to_model will be /my-directory
        ```

        ```python
            # Takes all the arguments passed to `snapshot_download`
            # except for `local_dir`
            @huggingface_hub(load=[
                {
                    "repo_id": "mistralai/Mistral-7B-Instruct-v0.1",
                },
                {
                    "repo_id": "myorg/mistral-lora",
                    "repo_type": "model",
                },
            ])
            @step
            def finetune_model(self):
                path_to_model = current.huggingface_hub.loaded["mistralai/Mistral-7B-Instruct-v0.1"]
                # path_to_model will be /my-directory
        ```
    """

    defaults = {
        "temp_dir_root": None,
        "load": None,  # Can be list of repo_ids or dicts with repo_id and other params
    }

    name = "huggingface_hub"

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        self._flow_datastore = flow_datastore
        self._logger = logger
        self._chkptr = None
        self._collector_thread = None

        self._registry = HuggingfaceRegistry(logger)

    def _resolve_settings(self):
        return {
            "load_policy": "none",
            "temp_dir_root": self.attributes.get("temp_dir_root"),
        }

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        super().task_pre_step(
            step_name,
            task_datastore,
            metadata,
            run_id,
            task_id,
            flow,
            graph,
            retry_count,
            max_user_code_retries,
            ubf_context,
            inputs,
        )
        self._runid, self._step_name, self._task_id = run_id, step_name, task_id
        self._metadata_provider = metadata
        # Handle loading models if load argument is provided
        load_models = self.attributes.get("load")
        if load_models is not None:
            if not isinstance(load_models, list):
                raise ValueError(
                    f"Invalid load argument format: {load_models}. "
                    "Must be a list of strings, tuples, or dicts"
                )

            for model_spec in load_models:
                if isinstance(model_spec, str):
                    # List[str] - repo_ids
                    self._registry.loaded._load_model(model_spec)
                elif isinstance(model_spec, tuple):
                    # Validate tuple length
                    if len(model_spec) != 2:
                        raise ValueError(
                            f"Invalid tuple format: {model_spec}. "
                            "Must be (dict, path) or (repo_id, path)"
                        )
                    spec, path = model_spec

                    if isinstance(spec, dict):
                        # List[Tuple[Dict, str]] - (model config, path)
                        if "repo_id" not in spec:
                            raise ValueError(
                                f"Missing repo_id in dict specification: {spec}"
                            )
                        repo_id = spec.pop("repo_id")
                        self._registry.loaded._load_model(repo_id, path=path, **spec)
                    elif isinstance(spec, str):
                        # List[Tuple[str, str]] - (repo_id, path)
                        self._registry.loaded._load_model(spec, path=path)
                    else:
                        raise ValueError(
                            f"Invalid tuple first element: {spec}. "
                            "Must be dict or string"
                        )
                elif isinstance(model_spec, dict):
                    # List[Dict] - model configs
                    if "repo_id" not in model_spec:
                        raise ValueError(
                            f"Missing repo_id in dict specification: {model_spec}"
                        )
                    repo_id = model_spec.pop("repo_id")
                    self._registry.loaded._load_model(repo_id, **model_spec)
                else:
                    raise ValueError(
                        f"Invalid model specification format: {model_spec}. "
                        "Must be string, tuple (dict/str, path), or dict"
                    )
        
        self.loaded_models_data = self._registry.loaded.info
        model_keys = [
            model_ref["key"] for _, model_ref in self.loaded_models_data.items()
        ]
        # Register metadata about the models that are loaded so that we can
        # use it for book keeping in the future.
        if len(model_keys) > 0:
            self._metadata_provider.register_metadata(
                self._runid,
                self._step_name,
                self._task_id,
                [
                    MetaDatum(
                        "hf-loaded-models",
                        json.dumps({"keys": model_keys}),
                        "model-registry",
                        tags=[
                            "attempt_id:%s" % str(retry_count),
                        ],
                    )
                ],
            )

    def _setup_current(self):
        from metaflow import current

        self._registry._set_checkpointer(self._chkptr)
        current._update_env({"huggingface_hub": self._registry})

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        return step_func

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        self._chkptr.cleanup()

    def task_exception(
        self, exception, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        self._chkptr.cleanup()



from typing import Union, Any


class SerializationHandler:

    TYPE = None

    STORAGE_FORMAT = None

    def serialze(self, *args, **kwargs) -> Union[str, bytes]:
        raise NotImplementedError

    def deserialize(self, *args, **kwargs) -> Any:
        raise NotImplementedError



from typing import Union
import tempfile
from ..tar_utils import (
    create_tarball_in_memory,
    create_tarball_on_disk,
    extract_tarball,
)
from .base import SerializationHandler


class TarHandler(SerializationHandler):

    TYPE = "tar"

    STORAGE_FORMAT = "tar"

    def serialize(
        self, path_to_compress, path_to_save=None, in_memory=False, strict=False
    ) -> Union[str, bytes]:
        if in_memory:
            return create_tarball_in_memory(
                path_to_compress, compression_method=None, strict=strict
            )
        else:
            # self._tempfile = tempfile.NamedTemporaryFile()
            create_tarball_on_disk(
                path_to_compress, path_to_save, compression_method=None, strict=strict
            )
            return path_to_save

    def deserialize(
        self,
        path_or_bytes,
        target_directory,
    ) -> str:
        extract_tarball(path_or_bytes, target_directory, compression_method=None)
        return target_directory



from metaflow.exception import MetaflowException


class CheckpointNotAvailableException(MetaflowException):
    headline = "Checkpoint not available"

    def __init__(self, message):
        super(CheckpointNotAvailableException, self).__init__(message)


class CheckpointException(MetaflowException):
    headline = "@checkpoint Exception"

    def __init__(self, message):
        super(CheckpointException, self).__init__(message)



from metaflow.exception import MetaflowException


class DatastoreReadInitException(MetaflowException):
    headline = "Exception initializing datastore for reading"

    def __init__(self, message):
        super(DatastoreReadInitException, self).__init__(message)


class DatastoreWriteInitException(MetaflowException):
    headline = "Exception initializing datastore for writing"

    def __init__(self, message):
        super(DatastoreWriteInitException, self).__init__(message)


class DatastoreNotReadyException(MetaflowException):
    headline = "Datastore not ready for write operations"

    def __init__(self, message):
        super(DatastoreNotReadyException, self).__init__(message)



import os
from collections import namedtuple
import json
import datetime

PathSize = namedtuple(
    "PathSize", ["path", "calculated_size", "early_stopping", "not_found"]
)
# calculated_size : size in bytes


def replace_start_and_end_slash(string):
    if string.startswith("/"):
        string = string[1:]
    if string.endswith("/"):
        string = string[:-1]
    return string


def get_path_size(start_path, early_stopping_limit=None) -> PathSize:
    total_size = 0
    if not os.path.exists(start_path):
        return PathSize(
            path=start_path, calculated_size=0, early_stopping=False, not_found=True
        )
    if os.path.isfile(start_path):
        return PathSize(
            path=start_path,
            calculated_size=os.path.getsize(start_path),
            early_stopping=False,
            not_found=False,
        )
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # Skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)
                if (
                    early_stopping_limit is not None
                    and total_size >= early_stopping_limit
                ):
                    return PathSize(
                        path=start_path,
                        calculated_size=total_size,
                        early_stopping=True,
                        not_found=False,
                    )

    return PathSize(
        path=start_path,
        calculated_size=total_size,
        early_stopping=False,
        not_found=False,
    )


def warning_message(message, logger=None, ts=False, prefix="[no-prefix]"):
    msg = "%s %s" % (prefix, message)
    if logger:
        logger(msg, timestamp=ts, bad=True)


def unit_convert(number, base_unit, convert_unit):
    # base_unit : GB or MB or KB or B
    # convert_unit : GB or MB or KB or B
    # number : number of base_unit
    # return : number of convert_unit
    units = ["B", "KB", "MB", "GB"]
    if base_unit not in units or convert_unit not in units:
        raise ValueError("Invalid unit")
    base_unit_index = units.index(base_unit)
    convert_unit_index = units.index(convert_unit)
    factor = pow(1024, abs(base_unit_index - convert_unit_index))
    if base_unit_index < convert_unit_index:
        return round(number / factor, 3)
    else:
        return round(number * factor, 3)


def is_json_serializable(value):
    """Check if value is JSON serializable."""
    try:
        json.dumps(value)
        return True
    except (TypeError, OverflowError):
        return False


def safe_serialize(data):
    def handle_value(val):
        """Converts non-serializable values to serializable format or a placeholder."""
        if isinstance(val, datetime.datetime):
            return val.isoformat()
        elif isinstance(val, (dict, list, tuple)):
            return safe_serialize(val)
        elif is_json_serializable(val):
            return val
        else:
            return "<value cannot be JSON serialized>"

    if isinstance(data, dict):
        return {key: handle_value(val) for key, val in data.items()}
    elif isinstance(data, list):
        return [handle_value(item) for item in data]
    elif isinstance(data, tuple):
        return tuple(handle_value(item) for item in data)
    else:
        return handle_value(data)



from typing import Iterable, Union, List, Dict, Any, Tuple, Optional, TYPE_CHECKING
import tempfile
from ..datastructures import CheckpointArtifact
from .constants import (
    CHECKPOINT_UID_ENV_VAR_NAME,
    DEFAULT_NAME,
    TASK_CHECKPOINTS_ARTIFACT_NAME,
    DEFAULT_STORAGE_FORMAT,
)
from .constructors import (
    _instantiate_checkpoint,
    load_checkpoint,
)
from .exceptions import CheckpointNotAvailableException

if TYPE_CHECKING:
    import metaflow
    from .core import Checkpointer


def _extract_checkpoints_for_task(
    task: Union["metaflow.Task", str], attempt: Optional[Union[int, str]] = None
):
    from metaflow import Task

    if isinstance(task, str):
        if attempt is not None:
            try:
                _attempt = int(attempt)
            except ValueError:
                raise ValueError("Attempt number must be an integer. Got: %s" % attempt)
            task = Task(task, attempt=_attempt, _namespace_check=False)
        task = Task(task, _namespace_check=False)
    try:
        return task[TASK_CHECKPOINTS_ARTIFACT_NAME].data
    except NameError:
        raise CheckpointNotAvailableException(
            "Checkpoints were not recorded for the task"
        )


class Checkpoint:

    _checkpointer: "Checkpointer" = None

    def __init__(self, temp_dir_root=None, init_dir=False):
        self._temp_dir_root = temp_dir_root
        self._checkpoint_dir = None
        if init_dir:
            self._checkpoint_dir = tempfile.TemporaryDirectory(dir=self._temp_dir_root)

    @property
    def directory(self):
        if self._checkpoint_dir is None:
            return None
        return self._checkpoint_dir.name

    def _set_checkpointer(self, checkpointer: "Checkpointer"):
        self._checkpointer = checkpointer

    def save(
        self,
        path=None,
        metadata=None,
        latest=True,
        name=DEFAULT_NAME,
        storage_format=DEFAULT_STORAGE_FORMAT,
    ):
        """
        Saves the checkpoint to the datastore

        Parameters
        ----------
        path : Optional[Union[str, os.PathLike]], default: None
            The path to save the checkpoint. Accepts a file path or a directory path.
                - If a directory path is provided, all the contents within that directory will be saved.
                When a checkpoint is reloaded during task retries, `the current.checkpoint.directory` will
                contain the contents of this directory.
                - If a file path is provided, the file will be directly saved to the datastore (with the same filename).
                When the checkpoint is reloaded during task retries, the file with the same name will be available in the
                `current.checkpoint.directory`.
                - If no path is provided then the `Checkpoint.directory` will be saved as the checkpoint.

        name : Optional[str], default: "mfchckpt"
            The name of the checkpoint.

        metadata : Optional[Dict], default: {}
            Any metadata that needs to be saved with the checkpoint.

        latest : bool, default: True
            If True, the checkpoint will be marked as the latest checkpoint.
            This helps determine if the checkpoint gets loaded when the task restarts.

        storage_format : str, default: files
            If `tar`, the contents of the directory will be tarred before saving to the datastore.
            If `files`, saves directory directly to the datastore.
        """
        if path is None and self.directory is None:
            raise ValueError(
                "`path` cannot be None when the Checkpoint object is not instantiated with a context manager. "
            )
        if path is None:
            path = self.directory
        if self._checkpointer is None:
            self = _instantiate_checkpoint(self)
        if metadata is None:
            metadata = {}
        return self._checkpointer.save(
            path=path,
            name=name,
            metadata=metadata,
            latest=latest,
            storage_format=storage_format,
        ).to_dict()

    def __enter__(self):
        if self._checkpoint_dir is None:
            self._checkpoint_dir = tempfile.TemporaryDirectory(dir=self._temp_dir_root)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._checkpoint_dir.cleanup()
        self._checkpoint_dir = None

    def list(
        self,
        name: Optional[str] = None,
        task: Optional[Union["metaflow.Task", str]] = None,
        attempt: Optional[Union[int, str]] = None,
        as_dict: bool = True,  # This is not public!
        within_task: bool = True,  # This is not public!
        # `within_task` is a hidden option. If set to False, it will list all checkpoints in "scope" (current namespace).
        # If true, it will list all checkpoints created within the task
    ) -> Iterable[Union[Dict, CheckpointArtifact]]:

        """
        lists the checkpoints in the datastore based on the Task.
        It will always be task scoped.

        Usage:
        ------

        ```python

        Checkpoint().list(name="best") # lists checkpoints in the current task with the name "best"
        Checkpoint().list(task="anotherflow/somerunid/somestep/sometask", name="best") # Identical as the above one but
        Checkpoint().list() # lists all the checkpoints in the current task

        ```

        Parameters
        ----------

        - `name`:
            - name of the checkpoint to filter for
        - `task`:
            - Task object outside the one that is currently set in the `Checkpoint` object; Can be a pathspec string.
        - `attempt`:
            - attempt number of the task (optional filter. If none, then lists all checkpoints from all attempts)
        """

        if self._checkpointer is None and task is None:
            raise ValueError(
                "Calling `Checkpoint.list` requires a `task` argument when its is called outside a Metaflow process."
            )
        _gen = None
        if task is not None:
            _gen = _extract_checkpoints_for_task(task, attempt)
            for chckpt in _gen:
                art = CheckpointArtifact.hydrate(chckpt)
                if name is not None and art.name != name:
                    continue
                if as_dict:
                    yield art.to_dict()
                else:
                    yield art
            return
        else:
            _gen = self._checkpointer._list_checkpoints(
                name=name, attempt=attempt, within_task=within_task
            )
        for checkpoint in _gen:
            if as_dict:
                yield checkpoint.to_dict()
            else:
                yield checkpoint

    def load(
        self,
        reference: Union[str, Dict, CheckpointArtifact],
        path: Optional[str] = None,
    ):
        """
        loads a checkpoint reference from the datastore. (resembles a read op)

        Parameters
        ----------

        `reference` :
            - can be a string, dict or a CheckpointArtifact object:
                - string: a string reference to the checkpoint (checkpoint key)
                - dict: a dictionary reference to the checkpoint
                - CheckpointArtifact: a CheckpointArtifact object reference to the checkpoint
        """
        if path is None and self.directory is None:
            raise ValueError(
                "`path` cannot be None when the Checkpoint object is not instantiated with a context manager. "
            )
        if path is None:
            path = self.directory

        load_checkpoint(
            checkpoint=reference,
            local_path=path,
        )



from metaflow.exception import MetaflowException
from metaflow.datastore.datastore_storage import DataStoreStorage
from metaflow.plugins.datastores.local_storage import LocalStorage
import os
from functools import partial
import sys
import shutil

from typing import Iterator, List, Union, Tuple, Optional
import os
import json
from pathlib import Path
from io import BytesIO
from collections import namedtuple
from datetime import datetime
import tempfile

from ..utils.tar_utils import create_tarball_on_disk, extract_tarball
from ..utils.general import safe_serialize
from .storage_injections import (
    STORAGE_INJECTIONS_SINGLE_FILE_SAVE,
    STORAGE_INJECTIONS_MULTIPLE_FILE_SAVE,
    STORAGE_INJECTIONS_LOAD_FILES,
)
from ..exceptions import KeyNotFoundError
import json

# mimic a subset of the behavior of the Metaflow S3Object
DatastoreBlob = namedtuple("DatastoreBlob", "blob url path")
ListPathResult = namedtuple("ListPathResult", "full_url key")
COMPRESSION_METHOD = None


class STORAGE_FORMATS:
    TAR = "tar"
    FILES = "files"


def warning_message(
    message, logger=None, ts=False, prefix="[@checkpoint][artifact-store]"
):
    msg = "%s %s" % (prefix, message)
    if logger:
        logger(msg, timestamp=ts, bad=True)
    else:
        print(msg, file=sys.stderr)


def allow_safe(func):
    def wrapper(*args, **kwargs):
        _safe = False
        if kwargs.pop("safe", False):
            _safe = True
        try:
            return func(*args, **kwargs)
        except KeyNotFoundError as e:
            if _safe:
                return None
            raise e
        except ValueError as e:
            if _safe:
                return None
            raise e

    return wrapper


def _extract_paths_inside_directory(directory):
    """
    This function will extract all the file-paths inside a directory
    and returns the following:
        - A list of relative paths of the files inside the directory
        - The total size of the directory in bytes.
        - The absolute paths of the files inside the directory.
    """
    import os

    relative_paths = []
    absolute_paths = []
    total_size = 0

    directory = os.path.abspath(directory)

    for root, dirs, files in os.walk(directory):
        for file in files:
            abs_path = os.path.join(root, file)
            rel_path = os.path.relpath(abs_path, directory)
            relative_paths.append(rel_path)
            absolute_paths.append(abs_path)
            total_size += os.path.getsize(abs_path)

    return relative_paths, total_size, absolute_paths


class ObjectStorage(object):
    """
    `ObjectStorage` wraps around the DataStoreStorage object and provides the lower level
    storage APIs needed by the subsequent classes. This datastore's main function that
    distingushes it from the DataStoreStorage object is that it manages a generalizable
    path structure over different storage backends (s3, azure, gs, local etc.).

    This object will be used to create multiple "Logical" datastores for constructs like
    `Checkpoints`, `Models` etc.

    Usage
    -----
    ```
    storage = ObjectStorage(
        storage_backend,
        root_prefix = "mf.checkpoints",
        path_components = ["artifacts", "MyFlow", "step_a", "cd2312rd12d", "x5avasdhtsdfqw"]
    )
    ```
    """

    # Prefix that is created using the root_prefix and the path_components
    FULL_PREFIX = None

    def __init__(
        self,
        storage_backend: DataStoreStorage,
        root_prefix: str,  # `mf.checkpoints`
        path_components: List,  # artifacts/flow_name/step_name/scope/task_identifier
    ):
        self._backend: DataStoreStorage = storage_backend
        self._storage_root = self.resolve_root(self._backend.TYPE)
        self._path_components = path_components
        self.set_full_prefix(root_prefix)
        self._inject_methods_to_storage_backend()

    @property
    def path_components(self):
        return self._path_components

    def _inject_methods_to_storage_backend(
        self,
    ):
        """
        The DataStoreStorage object is a core Metaflow object and
        it doesn't contain some a method for saving file. This method
        injects the save_file method into the DataStoreStorage object
        based on the storage backend.
        """

        if self._backend.TYPE not in STORAGE_INJECTIONS_SINGLE_FILE_SAVE:
            raise NotImplementedError(
                "Storage backend %s not supported with @checkpoint."
                % self._backend.TYPE
            )

        method = STORAGE_INJECTIONS_SINGLE_FILE_SAVE[self._backend.TYPE]
        setattr(self._backend, "save_file", partial(method, self._backend))

        if self._backend.TYPE not in STORAGE_INJECTIONS_MULTIPLE_FILE_SAVE:
            raise NotImplementedError(
                "Storage backend %s not supported with @checkpoint."
                % self._backend.TYPE
            )
        method = STORAGE_INJECTIONS_MULTIPLE_FILE_SAVE[self._backend.TYPE]
        setattr(self._backend, "save_files", partial(method, self._backend))

        if self._backend.TYPE not in STORAGE_INJECTIONS_LOAD_FILES:
            raise NotImplementedError(
                "Storage backend %s not supported with @checkpoint."
                % self._backend.TYPE
            )
        method = STORAGE_INJECTIONS_LOAD_FILES[self._backend.TYPE]
        setattr(self._backend, "load_files", partial(method, self._backend))

    def set_full_prefix(self, root_prefix):
        self.FULL_PREFIX = os.path.join(root_prefix, "/".join(self._path_components))

    @staticmethod
    def resolve_root(storage_type):
        if storage_type == "s3":
            from metaflow.metaflow_config import DATASTORE_SYSROOT_S3

            return DATASTORE_SYSROOT_S3
        elif storage_type == "azure":
            from metaflow.metaflow_config import DATASTORE_SYSROOT_AZURE

            return DATASTORE_SYSROOT_AZURE
        elif storage_type == "gs":
            from metaflow.metaflow_config import DATASTORE_SYSROOT_GS

            return DATASTORE_SYSROOT_GS
        elif storage_type == "local":
            return LocalStorage.get_datastore_root_from_config(
                lambda x: x, create_on_absent=True
            )
        else:
            raise NotImplementedError(
                "Datastore is not support backend %s" % (storage_type)
            )

    def full_base_url(self, prefix=None):
        if prefix is None:
            return self._storage_root
        return os.path.join(self._storage_root, prefix)

    def create_key_name(self, *args):
        return ".".join([str(arg) for arg in args])

    @property
    def datastore_root(self):
        assert self.FULL_PREFIX is not None, "FULL_PREFIX is not set."
        path_prefix = self.FULL_PREFIX
        # The FULL_PREFIX property is used to construct the root based on the
        # storage type and will be used to construct the path to the
        # objects in the datastore during read/write operations.

        # If the storage type is s3, then the root is just the `FULL_PREFIX`
        # because the S3Storage object uses as S3 client which has the
        # root already set.
        if self._backend.TYPE != "s3":
            path_prefix = self.full_base_url(prefix=path_prefix)

        return self.FULL_PREFIX

    def resolve_key_relative_path(self, key):
        return os.path.join(self.FULL_PREFIX, key)

    def resolve_key_full_url(self, key):
        return os.path.join(self.full_base_url(), self.resolve_key_relative_path(key))

    def resolve_key_path(self, key):
        return os.path.join(self.datastore_root, key)

    # TODO: [CORE-CLEANUP]: Make this HYPER Efficient!
    def put(
        self, key: str, obj: Union[str, bytes], overwrite: bool = False
    ) -> str:  # Path to where it got stored.
        """
        TODO : [THIS IS TERRIBLY INEFFICIENT]
        """
        "Put a single object into the datastore's `key` index."
        _save_object = None
        if isinstance(obj, bytes):
            _save_object = BytesIO(obj)
        else:
            _save_object = BytesIO(obj.encode("utf-8"))
        _path = self.resolve_key_path(key)
        self._backend.save_bytes(
            [(_path, _save_object)],
            overwrite=overwrite,
        )
        return self.resolve_key_relative_path(key)

    def put_file(self, key: str, path: str, overwrite=False):
        _kp = self.resolve_key_path(key)
        self._backend.save_file(_kp, path, overwrite=overwrite)
        return self.resolve_key_relative_path(key)

    def put_files(self, key_paths: List[Tuple[str, str]], overwrite=False):

        results = []
        for key, path in key_paths:
            _kp = self.resolve_key_path(key)
            self._backend.save_file(_kp, path, overwrite=overwrite)
            results.append(self.resolve_key_relative_path(key))
        return results

    # TODO: [CORE-CLEANUP]: Make this HYPER inefficient!
    def get(self, key) -> DatastoreBlob:
        """
        TODO : [THIS IS TERRIBLY INEFFICIENT]
        """
        "Get a single object residing in the datastore's `key` index."
        datastore_url = self.resolve_key_path(key)
        with self._backend.load_bytes([datastore_url]) as get_results:
            for key, path, meta in get_results:
                if path is not None:
                    with open(path, "rb") as f:
                        blob_bytes = f.read()
                        return DatastoreBlob(
                            blob=blob_bytes,
                            url=datastore_url,
                            path=path,
                            # text=blob_bytes.decode("utf-8"),
                        )
                else:
                    raise KeyNotFoundError(datastore_url)

    def get_file(self, key):
        datastore_url = self.resolve_key_path(key)
        return self._backend.load_bytes([datastore_url])

    def list_paths(self, keys, recursive=False) -> Iterator[ListPathResult]:
        "List all objects in the datastore's `keys` index."

        def _full_url_convert(lcr_path):
            if self._backend.TYPE == "s3":
                return os.path.join(self.full_base_url(), lcr_path)
            return lcr_path

        def _relative_url_convert(lcr_path):
            if self._backend.TYPE == "s3":
                return lcr_path
            np = lcr_path.replace(self.full_base_url(), "")
            if np.startswith("/"):
                return np[1:]
            return np

        keys = [self.resolve_key_path(key) for key in keys]
        if not recursive:
            for list_content_result in self._backend.list_content(keys):
                yield ListPathResult(
                    full_url=_full_url_convert(list_content_result.path),
                    key=_relative_url_convert(list_content_result.path),
                )
        else:

            def _list_content_recursive(x_keys):
                _keys = []
                for list_content_result in self._backend.list_content(x_keys):
                    if list_content_result.is_file:
                        _keys.append(
                            ListPathResult(
                                full_url=_full_url_convert(list_content_result.path),
                                key=_relative_url_convert(list_content_result.path),
                            )
                        )
                    else:
                        _keys.extend(
                            _list_content_recursive([list_content_result.path])
                        )
                return _keys

            for x in _list_content_recursive(keys):
                yield x

    def _save_objects(
        self,
        key: str,
        local_path: str,
    ):
        # this function will just store all the objects as is to the datastore without taring them
        # up. This is useful when we are storing a single file or a directory that doesn't need to be
        # tarred up.
        object_size = None
        relative_paths_inside_key = []
        absolute_paths_of_files = []
        if os.path.isdir(local_path):
            (
                relative_paths_inside_key,
                object_size,
                absolute_paths_of_files,
            ) = _extract_paths_inside_directory(local_path)
        else:
            object_size = os.path.getsize(local_path)
            relative_paths_inside_key.append(
                os.path.basename(local_path),
            )
            absolute_paths_of_files.append(local_path)

        # at this point, since we already have a key, we need to ensure that we are placing everything
        # under the key. This is because the key will be used as the identifier and we will end up extracting
        # the contents of the object inside the key to the local path.
        key_paths = [
            (self.resolve_key_path(os.path.join(key, rel_path)), abs_path)
            for rel_path, abs_path in zip(
                relative_paths_inside_key, absolute_paths_of_files
            )
        ]
        self._backend.save_files(key_paths, overwrite=True)
        return (
            self.resolve_key_full_url(key),
            self.resolve_key_relative_path(key),
            object_size,
        )

    def _save_tarball(
        self,
        key,
        local_path: str,
    ):
        suffix = ".tar"
        with tempfile.NamedTemporaryFile(suffix=suffix) as temp_file:
            create_tarball_on_disk(
                local_path,
                output_filename=temp_file.name,
                compression_method=None,
            )
            file_size = os.path.getsize(temp_file.name)
            _ = self.put_file(key, temp_file.name, overwrite=True)
            return (
                self.resolve_key_full_url(key),
                self.resolve_key_relative_path(key),
                file_size,
            )

    def _load_objects(
        self,
        key,
        # The key here is ideally a key from the datastore that we want to load
        # This can be a checkpoint key or a model key.
        local_directory,
        # Its assumed here that the local path where everything is getting
        # extracted is a directory.
    ):
        list_path_results = list(self.list_paths([key], recursive=True))
        # print(list_path_results)
        keys = [p.key for p in list_path_results]
        # We directly call load bytes here because `self.get_file` will add the root of the datastore
        # to the path and we don't want that.
        with self._backend.load_files(keys) as get_results:
            for list_key, path, meta in get_results:
                if path is None:
                    continue

                path_within_dir = os.path.relpath(list_key, self.resolve_key_path(key))
                # We need to construct the right path over here based on the
                # where the key is present in the object.
                # Figure a relative path from the end of the key
                # to the actual file/directory within it.
                # We do this because we want the entire directory structure
                # to be preserved when we download the objects on local.
                move_to_path = os.path.join(local_directory, path_within_dir)
                Path(move_to_path).parent.mkdir(parents=True, exist_ok=True)
                shutil.move(
                    path,
                    move_to_path,
                )

        return local_directory

    def _load_tarball(
        self,
        key,
        local_path,
    ):

        with self.get_file(key) as get_results:
            for key, path, meta in get_results:
                if path is None:
                    raise KeyNotFoundError(key)
                extract_tarball(path, local_path, compression_method=None)
        return local_path

    def _save_metadata(
        self,
        key,
        metadata,
    ):
        return self.put(key, json.dumps(safe_serialize(metadata)), overwrite=True)

    def _load_metadata(self, key):
        return json.loads(self.get(key).blob)


class DatastoreInterface:
    """
    This is the root abstraction used by any underlying datastores like Checkpoint/Model etc.
    to create the saving/loading mechanics using multiple ObjectStores.

    The inherited classes require the following implemented:
        - `ROOT_PREFIX` : The root prefix for the datastore such as `mf.checkpoints` or `mf.models`.
        - `init_read_store` : The method to initialize the read store; The inheriting class can compose together any number of `BaseDatastore` objects
        - `init_write_store` : The method to initialize the write store; The inheriting class can compose together any number of `BaseDatastore` objects
        - `save` : The method to save the artifact data.
        - `load` : The method to load the artifact data.
        - `save_metadata` : The method to save the metadata about the artifact.
        - `load_metadata` : The method to load the metadata about the artifact.
        - `list` : The method to list all the artifacts.
    """

    ROOT_PREFIX = None

    def set_root_prefix(self, root_prefix):
        """
        This function helps ensuring that the root prefix of the datastore
        and it's underlying ObjectStores can change trivially.
        """
        raise NotImplementedError

    # ! INHERIT AND OVERRIDE
    def save(self, *args, **kwargs):
        raise NotImplementedError

    # ! INHERIT AND OVERRIDE
    def load(self, *args, **kwargs):
        raise NotImplementedError

    # ! INHERIT AND OVERRIDE
    def save_metadata(self, *args, **kwargs):
        raise NotImplementedError

    # ! INHERIT AND OVERRIDE
    def load_metadata(self, *args, **kwargs):
        raise NotImplementedError

    # ! INHERIT AND OVERRIDE
    def list(self, *args, **kwargs):
        raise NotImplementedError

    # ! INHERIT AND OVERRIDE
    @classmethod
    def init_read_store(cls, storage_backend: DataStoreStorage, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def init_write_store(cls, storage_backend: DataStoreStorage, *args, **kwargs):
        raise NotImplementedError


def resolve_root(datastore_type):
    return ObjectStorage.resolve_root(datastore_type)



import json
import time
from typing import List, Union
from ...card_utils import (
    CardDecoratorInjector,
    CardRefresher,
    AsyncPeriodicRefresher,
    LineChart,
    UpadateableTable,
)
from ...datastructures import CheckpointArtifact
from .lineage_card import (
    construct_lineage_table,
    create_checkpoint_card,
    null_card,
    format_datetime,
)
from ...utils.general import unit_convert
from metaflow.cards import Markdown, Table, Artifact, VegaChart
from datetime import datetime, timedelta
from threading import Thread, Event


import json
from datetime import datetime


def human_readable_date(date):
    return date.strftime("%Y-%m-%d %H:%M:%S")


def determine_nice_value(time_range_seconds):
    """
    Function to determine the 'nice' value based on the time range in seconds.
    """
    if time_range_seconds <= 60 * 30:  # less than 30 mins
        return "second", "%H:%M:%S"
    elif time_range_seconds <= 3600:  # less than an hour
        return "minute", "%H:%M"
    elif time_range_seconds <= 86400:  # less than a day
        return "hour", "%H:%M"
    elif time_range_seconds <= 604800:  # less than a week
        return "day", "%Y-%m-%d"
    elif time_range_seconds <= 2592000:  # less than a month
        return "week", "%Y-%m-%d"
    elif time_range_seconds <= 31536000:  # less than a year
        return "month", "%Y-%m"
    else:
        return "year", "%Y"


def generate_vega_timeline(data_objects):
    # Parse the created_on field into datetime objects
    for obj in data_objects:
        obj["created_on"] = datetime.fromisoformat(obj["created_on"])

    # Sort the data by creation time
    sorted_data = sorted(data_objects, key=lambda x: x["created_on"])

    # Get the earliest and latest timestamps
    earliest_time = sorted_data[0]["created_on"]
    latest_time = sorted_data[-1]["created_on"]

    # Calculate the time range in seconds
    time_range_seconds = (latest_time - earliest_time).total_seconds()

    # If the time range is too small, add some padding to make the domain larger
    if time_range_seconds < 10:
        earliest_time = earliest_time - timedelta(seconds=10)
        latest_time = latest_time + timedelta(seconds=10)

    # Convert to milliseconds since epoch for Vega
    min_t = earliest_time.timestamp() * 1000
    max_t = latest_time.timestamp() * 1000

    _CHART_HEIGHT = 250
    _HEIGHT_SCALE = 50

    # Determine the appropriate 'nice' value based on the time range
    nice_value, date_format_for_x_axis = determine_nice_value(time_range_seconds)

    # We use a dynamic scaling factor to ensure that the rectangles in the timeline chart
    # don't overlap thereby we need to extend the horizontal line under each rectangle dynamically
    # based on when was the previous
    names = list(set(x["name"] for x in sorted_data))
    max_bars = _CHART_HEIGHT // _HEIGHT_SCALE
    bar_heights_by_type = {}
    for idx, name in enumerate(names):
        _scaling_factor = idx % max_bars
        bar_heights_by_type[name] = (
            _scaling_factor,
            _HEIGHT_SCALE + (_scaling_factor * _HEIGHT_SCALE),
        )

    for obj in sorted_data:
        obj["bar_height"] = bar_heights_by_type[obj["name"]][1]
        obj["scaling_factor"] = bar_heights_by_type[obj["name"]][0]
    # Prepare the Vega data array for the timeline
    vega_data = [
        {
            "name_version": obj["version_id"],
            "name": obj["name"],
            "created_on": obj[
                "created_on"
            ].isoformat(),  # ISO format is directly supported by Vega
            "timestamp": obj["created_on"].timestamp()
            * 1000,  # Timestamps in milliseconds
            "bar_height": obj["bar_height"],
            "description": f"Version ID: {obj['name']}.{obj['version_id']}, Created On: {human_readable_date(obj['created_on'])}",
        }
        for obj in sorted_data
    ]

    # Vega spec
    vega_spec = {
        "$schema": "https://vega.github.io/schema/vega/v5.json",
        "width": 800,  # Increased the width
        "height": _CHART_HEIGHT,
        "padding": 5,
        "config": {
            "text": {"font": "Helvetica, Arial, sans-serif"},
            "axis": {"labelFont": "Helvetica, Arial, sans-serif", "labelFontSize": 10},
        },
        "signals": [
            {"name": "rectWidth", "value": 40},
            {"name": "rectHeight", "value": 30},
            {"name": "rectY", "value": 55},
            {"name": "rectCenter", "init": "[rectWidth/2,rectY+rectHeight/2]"},
        ],
        "data": [
            {
                "name": "checkpoints",
                "format": {"type": "json"},
                "values": vega_data,
                "transform": [
                    {
                        "type": "collect",
                        "sort": {"field": "timestamp"},  # Use timestamp for sorting
                    }
                ],
            }
        ],
        "scales": [
            {
                "name": "xScale",
                "type": "time",  # Use time scale for proper date spacing
                "domain": [min_t, max_t],  # Use min_t and max_t for the domain
                "range": [
                    {"signal": "0"},
                    {"signal": "width"},
                ],  # Correct range using width signal
                "nice": nice_value,  # Dynamically set the 'nice' value
            },
            {
                "name": "colorScale",
                "type": "ordinal",
                "domain": {"data": "checkpoints", "field": "name"},
                "range": {"scheme": "category10"},  # Use a categorical color scheme
            },
        ],
        "axes": [
            {
                "scale": "xScale",
                "orient": "bottom",
                "format": date_format_for_x_axis,
                "labelOverlap": True,
                "labelAngle": 0,  # Angled labels for better readability
            }
        ],
        "legends": [
            {
                "fill": "colorScale",
                "title": "Checkpoints",
                "orient": "right",
                "labelFontSize": 12,
                "titleFontSize": 14,
            }
        ],
        "marks": [
            {
                "type": "rect",
                "name": "rectangles",
                "zindex": 2,
                "from": {"data": "checkpoints"},
                "encode": {
                    "enter": {
                        "width": {"signal": "rectWidth"},
                        "height": {"signal": "rectHeight"},
                        "x": {"signal": "scale('xScale',datum.timestamp)-rectWidth/2"},
                        "y": {"signal": "datum.bar_height"},
                        "rectCenter": {"signal": "[rectWidth / 2, datum.bar_height]"},
                        "fill": {"scale": "colorScale", "field": "name"},
                        "tooltip": {"signal": "{'Description': datum.description}"},
                        "bar_height": {"field": "bar_height"},
                    },
                    "update": {"fillOpacity": {"value": 1}},
                    "hover": {"fillOpacity": {"value": 0.5}},
                },
            },
            {
                "type": "text",
                "name": "labels",
                "zindex": 3,
                "from": {"data": "rectangles"},
                "encode": {
                    "enter": {
                        "text": {"signal": "datum.datum.name_version"},
                        "x": {"signal": "datum.x+datum.rectCenter[0]"},
                        "y": {"signal": "datum.rectCenter[1] + rectHeight/2"},
                        "align": {"value": "center"},
                        "rule_y": {"signal": "datum.rectCenter[1]"},
                        # "rectCenter": {"field": "datum.rectCenter"},
                        "baseline": {"value": "middle"},
                        "fontWeight": {"value": "bold"},
                        "fill": {"value": "black"},
                        "bar_height": {"field": "datum.bar_height"},
                    }
                },
                "interactive": False,
            },
            {
                "type": "rule",
                "from": {"data": "labels"},
                "encode": {
                    "enter": {
                        "x": {"signal": "datum.x"},
                        "x2": {"signal": "datum.x"},
                        "y": {"signal": "datum.rule_y+rectHeight"},
                        "y2": {"signal": "height"},
                        "strokeWidth": {"value": 2},
                    }
                },
            },
        ],
    }

    return vega_spec


class CheckpointListRefresher(CardRefresher):

    CARD_ID = "task_checkpoints"

    TABLE_HEADERS = [
        "Name",
        "Created On",
        "Size",
        "Metadata",
        # "Key",
    ]

    def __init__(
        self,
        loaded_checkpoint: Union[CheckpointArtifact, None],
        lineage_stack: Union[List[CheckpointArtifact], None],
        load_policy: str,
    ) -> None:
        self._rendered = False
        self._errored = False
        self._table = None
        self._timeline_chart = None
        self._loaded_checkpoint = loaded_checkpoint
        self._lineage_stack = lineage_stack
        self._load_policy = load_policy

        from metaflow import current

        self.current = current
        self._saved_checkpoints = {}

    def on_error(self, current_card, error_message):
        if isinstance(error_message, FileNotFoundError):
            return
        if isinstance(error_message, json.JSONDecodeError):
            return
        current_card.clear()
        current_card.append(
            Markdown(
                f"## Error: {str(error_message)}",
            )
        )
        current_card.refresh()
        self._errored = True
        self._rendered = False

    def _header_components(self):
        x = [
            Markdown(
                "# Checkpoints \n **Task %s [Attempt:%s]**"
                % (self.current.pathspec, self.current.retry_count),
            )
        ]
        if self._loaded_checkpoint is not None:
            x.extend(
                create_checkpoint_card(
                    self._loaded_checkpoint, self._lineage_stack, self._load_policy
                )
            )
        else:
            x.extend(null_card(self._load_policy))
        return x

    def _footer_components(self):
        if not self._loaded_checkpoint:
            return [
                Markdown("## Lineage of Loaded Checkpoint"),
                Markdown(
                    "_no lineage found_",
                ),
            ]
        lineage_md = Markdown("## Lineage of Loaded Checkpoint")
        lineage_table = construct_lineage_table(self._lineage_stack)
        return [lineage_md, lineage_table]

    def on_startup(self, current_card):
        current_card.extend(self._header_components())
        current_card.extend(self._footer_components())
        current_card.refresh()

    def first_time_render(self, current_card, data_object, force_refresh=False):
        current_card.clear()
        current_card.extend(self._header_components())
        keys_going_in_table = self._make_table_objects(data_object)
        if len(keys_going_in_table) == 0:
            current_card.extend(
                [
                    Markdown("## Checkpoints created within the task"),
                    Markdown(
                        "_no checkpoints found_",
                    ),
                ]
            )
            current_card.extend(self._footer_components())
            current_card.refresh()
            return

        self._table = UpadateableTable(
            data=[self._saved_checkpoints[key] for key in keys_going_in_table],
            headers=self.TABLE_HEADERS,
        )

        self._timeline_chart = VegaChart(
            generate_vega_timeline(data_object), show_controls=True
        )

        current_card.append(Markdown("## Checkpoints Timeline"))
        current_card.append(self._timeline_chart)
        current_card.append(Markdown("## Checkpoints created within the task"))
        current_card.append(self._table)
        current_card.extend(self._footer_components())
        current_card.refresh(force=force_refresh)
        self._rendered = True

    def _make_table_objects(self, data_object):
        keys_going_in_table = []
        for checkpoint in data_object:
            _chckpt = CheckpointArtifact.from_dict(checkpoint)
            if _chckpt.key in self._saved_checkpoints:
                continue
            self._saved_checkpoints[_chckpt.key] = [
                Markdown(str(_chckpt.name)),
                Markdown(format_datetime(str(_chckpt.created_on))),
                Markdown(
                    _derive_appropriate_size(_chckpt.size),
                ),
                Artifact(_chckpt.metadata),
                # Markdown("```json\n%s\n```" % json.dumps(_chckpt.metadata, indent=4)),
                # Markdown(_chckpt.key),
            ]
            keys_going_in_table.append(_chckpt.key)
        return keys_going_in_table

    def data_update(self, current_card, data_object):
        keys_going_in_table = self._make_table_objects(data_object)
        if len(keys_going_in_table) == 0:
            return
        for key in keys_going_in_table:
            self._table.update(self._saved_checkpoints[key])

        self._timeline_chart.update(generate_vega_timeline(data_object))

        current_card.refresh()

    def on_update(self, current_card, data_object):
        if not self._rendered:
            self.first_time_render(current_card, data_object, force_refresh=False)
        else:
            self.data_update(current_card, data_object)

    def on_final(self, current_card, data_object):
        self._saved_checkpoints = {}
        self._rendered = False
        self.first_time_render(current_card, data_object, force_refresh=True)


def _derive_appropriate_size(size):
    if unit_convert(size, "B", "MB") < 1:
        return "%s KB" % str(unit_convert(size, "B", "KB"))
    elif unit_convert(size, "B", "GB") < 1:
        return "%s MB" % str(unit_convert(size, "B", "MB"))
    else:
        return "%s GB" % str(unit_convert(size, "B", "GB"))


class CheckpointsCollector(Thread):
    def __init__(self, refresher: CardRefresher, interval=1):
        super().__init__()
        from metaflow import current

        self.current = current
        self._interval = interval
        self._exit_event = Event()
        self._refresher = refresher

    def collect(self):
        return list(self.current.checkpoint.list(attempt=self.current.retry_count))

    def final_update(self):
        current_card = self.current.card[self._refresher.CARD_ID]
        data = self.collect()
        if len(data) == 0:
            return
        self._refresher.on_final(current_card, data)

    def run_update(self):
        current_card = self.current.card[self._refresher.CARD_ID]
        data = self.collect()
        if len(data) == 0:
            return
        self._refresher.on_update(current_card, data)

    def run(self):
        if self._refresher.CARD_ID is None:
            raise ValueError("CARD_ID must be defined")
        current_card = self.current.card[self._refresher.CARD_ID]
        self._refresher.on_startup(current_card)
        while self._exit_event.is_set() is False:
            self.run_update()
            time.sleep(self._interval)

    def stop(self):
        if not self._exit_event.is_set():
            self._exit_event.set()
            # We expose a `final_update` so that the card can be
            # called with a `force` update so that the new card
            # is rendered when the thread is stopped.
            self.final_update()
            self.join()
